/* -*- C++ -*-
 *
 * This file is a part of LEMON, a generic C++ optimization library
 *
 * Copyright (C) 2003-2006
 * Egervary Jeno Kombinatorikus Optimalizalasi Kutatocsoport
 * (Egervary Research Group on Combinatorial Optimization, EGRES).
 *
 * Permission to use, modify and distribute this software is granted
 * provided that this copyright notice appears in all copies. For
 * precise terms see the accompanying LICENSE file.
 *
 * This software is provided "AS IS" with no warranty of any kind,
 * express or implied, and with no claim as to its suitability for any
 * purpose.
 *
 */

#ifndef LEMON_FULL_GRAPH_H
#define LEMON_FULL_GRAPH_H

#include <cmath>

#include <lemon/bits/graph_extender.h>

#include <lemon/bits/invalid.h>
#include <lemon/bits/utility.h>


///\ingroup graphs
///\file
///\brief FullGraph class.


namespace lemon {

  /// \brief Base of the FullGrpah.
  ///
  /// Base of the FullGrpah.
  class FullGraphBase {
    int _nodeNum;
    int _edgeNum;
  public:

    typedef FullGraphBase Graph;

    class Node;
    class Edge;

  public:

    FullGraphBase() {}


    ///Creates a full graph with \c n nodes.
    void construct(int n) { _nodeNum = n; _edgeNum = n * n; }
    
    typedef True NodeNumTag;
    typedef True EdgeNumTag;

    /// \brief Returns the node with the given index.
    ///
    /// Returns the node with the given index. Because it is a
    /// static size graph the node's of the graph can be indiced
    /// by the range from 0 to \e nodeNum()-1 and the index of
    /// the node can accessed by the \e index() member.
    Node operator()(int index) const { return Node(index); }

    /// \brief Returns the index of the node.
    ///
    /// Returns the index of the node. Because it is a
    /// static size graph the node's of the graph can be indiced
    /// by the range from 0 to \e nodeNum()-1 and the index of
    /// the node can accessed by the \e index() member.
    int index(const Node& node) const { return node.id; }

    ///Number of nodes.
    int nodeNum() const { return _nodeNum; }
    ///Number of edges.
    int edgeNum() const { return _edgeNum; }

    /// Maximum node ID.
    
    /// Maximum node ID.
    ///\sa id(Node)
    int maxNodeId() const { return _nodeNum-1; }
    /// Maximum edge ID.
    
    /// Maximum edge ID.
    ///\sa id(Edge)
    int maxEdgeId() const { return _edgeNum-1; }

    Node source(Edge e) const { return e.id % _nodeNum; }
    Node target(Edge e) const { return e.id / _nodeNum; }


    /// Node ID.
    
    /// The ID of a valid Node is a nonnegative integer not greater than
    /// \ref maxNodeId(). The range of the ID's is not surely continuous
    /// and the greatest node ID can be actually less then \ref maxNodeId().
    ///
    /// The ID of the \ref INVALID node is -1.
    ///\return The ID of the node \c v. 

    static int id(Node v) { return v.id; }
    /// Edge ID.
    
    /// The ID of a valid Edge is a nonnegative integer not greater than
    /// \ref maxEdgeId(). The range of the ID's is not surely continuous
    /// and the greatest edge ID can be actually less then \ref maxEdgeId().
    ///
    /// The ID of the \ref INVALID edge is -1.
    ///\return The ID of the edge \c e. 
    static int id(Edge e) { return e.id; }

    static Node nodeFromId(int id) { return Node(id);}
    
    static Edge edgeFromId(int id) { return Edge(id);}

    typedef True FindEdgeTag;

    /// Finds an edge between two nodes.
    
    /// Finds an edge from node \c u to node \c v.
    ///
    /// If \c prev is \ref INVALID (this is the default value), then
    /// It finds the first edge from \c u to \c v. Otherwise it looks for
    /// the next edge from \c u to \c v after \c prev.
    /// \return The found edge or INVALID if there is no such an edge.
    Edge findEdge(Node u,Node v, Edge prev = INVALID) const {
      return prev.id == -1 ? Edge(*this, u.id, v.id) : INVALID;
    }
    
      
    class Node {
      friend class FullGraphBase;

    protected:
      int id;
      Node(int _id) : id(_id) {}
    public:
      Node() {}
      Node (Invalid) : id(-1) {}
      bool operator==(const Node node) const {return id == node.id;}
      bool operator!=(const Node node) const {return id != node.id;}
      bool operator<(const Node node) const {return id < node.id;}
    };
    


    class Edge {
      friend class FullGraphBase;
      
    protected:
      int id;  // _nodeNum * target + source;

      Edge(int _id) : id(_id) {}

      Edge(const FullGraphBase& _graph, int source, int target) 
	: id(_graph._nodeNum * target+source) {}
    public:
      Edge() { }
      Edge (Invalid) { id = -1; }
      bool operator==(const Edge edge) const {return id == edge.id;}
      bool operator!=(const Edge edge) const {return id != edge.id;}
      bool operator<(const Edge edge) const {return id < edge.id;}
    };

    void first(Node& node) const {
      node.id = _nodeNum-1;
    }

    static void next(Node& node) {
      --node.id;
    }

    void first(Edge& edge) const {
      edge.id = _edgeNum-1;
    }

    static void next(Edge& edge) {
      --edge.id;
    }

    void firstOut(Edge& edge, const Node& node) const {
      edge.id = _edgeNum + node.id - _nodeNum;
    }

    void nextOut(Edge& edge) const {
      edge.id -= _nodeNum;
      if (edge.id < 0) edge.id = -1;
    }

    void firstIn(Edge& edge, const Node& node) const {
      edge.id = node.id * _nodeNum;
    }
    
    void nextIn(Edge& edge) const {
      ++edge.id;
      if (edge.id % _nodeNum == 0) edge.id = -1;
    }

  };

  typedef GraphExtender<FullGraphBase> ExtendedFullGraphBase;

  /// \ingroup graphs
  ///
  /// \brief A full graph class.
  ///
  /// This is a simple and fast directed full graph implementation.
  /// It is completely static, so you can neither add nor delete either
  /// edges or nodes.
  /// Thus it conforms to
  /// the \ref concept::Graph "Graph" concept
  /// \sa concept::Graph.
  ///
  /// \sa FullGraphBase
  /// \sa FullUGraph
  ///
  /// \author Alpar Juttner
  class FullGraph : public ExtendedFullGraphBase {
  public:

    typedef ExtendedFullGraphBase Parent;

    /// \brief Constructor
    FullGraph() { construct(0); }

    /// \brief Constructor
    ///
    FullGraph(int n) { construct(n); }

    /// \brief Resize the graph
    ///
    /// Resize the graph. The function will fully destroy and build the graph.
    /// This cause that the maps of the graph will reallocated
    /// automatically and the previous values will be lost.
    void resize(int n) {
      Parent::getNotifier(Edge()).clear();
      Parent::getNotifier(Node()).clear();
      construct(n);
      Parent::getNotifier(Node()).build();
      Parent::getNotifier(Edge()).build();
    }
  };

} //namespace lemon


#endif //LEMON_FULL_GRAPH_H
